# Copyright 2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import base64
from nats.js import api
from nats.errors import *
from nats.js.errors import *
from nats.aio.errors import *
from email.parser import BytesParser
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional

NATS_HDR_LINE = bytearray(b'NATS/1.0\r\n')
NATS_HDR_LINE_SIZE = len(NATS_HDR_LINE)


class JetStreamManager:
    """
    JetStreamManager exposes management APIs for JetStream.
    """
    def __init__(
        self,
        conn,
        prefix=api.DefaultPrefix,
        domain=None,
        timeout=5,
    ):
        self._prefix = prefix
        if domain is not None:
            self._prefix = f"$JS.{domain}.API"
        self._nc = conn
        self._timeout = timeout
        self._hdr_parser = BytesParser()

    async def account_info(self):
        info = await self._api_request(
            f"{self._prefix}.INFO", b'', timeout=self._timeout
        )
        return api.AccountInfo.loads(**info)

    async def find_stream_name_by_subject(self, subject: str):
        """
        Find the stream to which a subject belongs in an account.
        """

        req_sub = f"{self._prefix}.STREAM.NAMES"
        req_data = json.dumps({"subject": subject})
        info = await self._api_request(
            req_sub, req_data.encode(), timeout=self._timeout
        )
        return info['streams'][0]

    async def stream_info(self, name):
        """
        Get the latest StreamInfo by stream name.
        """
        resp = await self._api_request(
            f"{self._prefix}.STREAM.INFO.{name}", timeout=self._timeout
        )
        return api.StreamInfo.loads(**resp)

    async def add_stream(self, config: api.StreamConfig = None, **params):
        """
        add_stream creates a stream.
        """
        if not config:
            config = api.StreamConfig.loads(**params)

        if config.name is None:
            raise ValueError("nats: stream name is required")

        data = config.asjson()
        resp = await self._api_request(
            f"{self._prefix}.STREAM.CREATE.{config.name}",
            data.encode(),
            timeout=self._timeout
        )
        return api.StreamInfo.loads(**resp)

    async def delete_stream(self, name):
        """
        Delete a stream by name.
        """
        resp = await self._api_request(
            f"{self._prefix}.STREAM.DELETE.{name}", timeout=self._timeout
        )
        return resp['success']

    async def consumer_info(self, stream, consumer, timeout=None):
        # TODO: Validate the stream and consumer names.
        msg = None
        if timeout is None:
            timeout = self._timeout
        resp = await self._api_request(
            f"{self._prefix}.CONSUMER.INFO.{stream}.{consumer}",
            b'',
            timeout=timeout
        )
        return api.ConsumerInfo.loads(**resp)

    async def add_consumer(
        self,
        stream: str,
        config: Optional[api.ConsumerConfig] = None,
        durable_name: Optional[str] = None,
        description: Optional[str] = None,
        deliver_subject: Optional[str] = None,
        deliver_group: Optional[str] = None,
        deliver_policy: Optional[api.DeliverPolicy] = api.DeliverPolicy.last,
        opt_start_seq: Optional[int] = None,
        opt_start_time: Optional[int] = None,
        ack_policy: Optional[api.AckPolicy] = api.AckPolicy.explicit,
        ack_wait: Optional[int] = None,
        max_deliver: Optional[int] = None,
        filter_subject: Optional[str] = None,
        replay_policy: Optional[api.ReplayPolicy] = api.ReplayPolicy.instant,
        sample_freq: Optional[str] = None,
        rate_limit_bps: Optional[int] = None,
        max_waiting: Optional[int] = None,
        max_ack_pending: Optional[int] = None,
        flow_control: Optional[bool] = None,
        idle_heartbeat: Optional[int] = None,
        headers_only: Optional[bool] = None,
        timeout=None,
    ):
        if not timeout:
            timeout = self._timeout

        if config is None:
            config = {
                "durable_name": durable_name,
                "description": description,
                "deliver_subject": deliver_subject,
                "deliver_group": deliver_group,
                "deliver_policy": deliver_policy,
                "opt_start_seq": opt_start_seq,
                "opt_start_time": opt_start_time,
                "ack_policy": ack_policy,
                "ack_wait": ack_wait,
                "max_deliver": max_deliver,
                "filter_subject": filter_subject,
                "replay_policy": replay_policy,
                "max_waiting": max_waiting,
                "max_ack_pending": max_ack_pending,
                "flow_control": flow_control,
                "idle_heartbeat": idle_heartbeat,
                "headers_only": headers_only,
            }
        elif isinstance(config, api.ConsumerConfig):
            # Try to transform the config.
            durable_name = config.durable_name
            config = asdict(config)

        # Cleanup empty values.
        # FIXME: Make this part of Base
        for k, v in dict(config).items():
            if v is None:
                del config[k]
        req = {"stream_name": stream, "config": config}
        req_data = json.dumps(req).encode()

        resp = None
        if durable_name is not None:
            resp = await self._api_request(
                f"{self._prefix}.CONSUMER.DURABLE.CREATE.{stream}.{durable_name}",
                req_data,
                timeout=timeout
            )
        else:
            resp = await self._api_request(
                f"{self._prefix}.CONSUMER.CREATE.{stream}",
                req_data,
                timeout=timeout
            )
        return api.ConsumerInfo.loads(**resp)

    async def delete_consumer(self, stream=None, consumer=None):
        resp = await self._api_request(
            f"{self._prefix}.CONSUMER.DELETE.{stream}.{consumer}",
            b'',
            timeout=self._timeout
        )
        return resp['success']

    async def get_last_msg(
        self,
        stream_name: str,
        subject: str,
    ):
        """
        get_last_msg retrieves a message from a stream.
        """
        req_subject = f"{self._prefix}.STREAM.MSG.GET.{stream_name}"
        req = {'last_by_subj': subject}
        data = json.dumps(req)
        resp = await self._api_request(req_subject, data.encode())
        raw_msg = api.RawStreamMsg.loads(**resp['message'])
        if raw_msg.hdrs:
            hdrs = base64.b64decode(raw_msg.hdrs)
            raw_headers = hdrs[len(NATS_HDR_LINE):]
            parsed_headers = self._jsm._hdr_parser.parsebytes(raw_headers)
            headers = {}
            for k, v in parsed_headers.items():
                headers[k] = v
            raw_msg.headers = headers
        return raw_msg

    async def _api_request(self, req_subject, req=b'', timeout=5):
        resp = None
        try:
            msg = await self._nc.request(req_subject, req, timeout=timeout)
            resp = json.loads(msg.data)
        except NoRespondersError:
            raise ServiceUnavailableError

        # Check for API errors.
        if 'error' in resp:
            raise APIError.from_error(resp['error'])

        return resp



# Copyright 2016-2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
from dataclasses import dataclass, field


class Msg:
    """
    Msg represents a message delivered by NATS.
    """
    __slots__ = (
        'subject', 'reply', 'data', 'sid', '_client', 'headers', '_metadata'
    )

    class Ack:
        Ack = b"+ACK"
        Nak = b"-NAK"
        Progress = b"+WPI"
        Term = b"+TERM"

        # Reply metadata...
        Prefix0 = '$JS'
        Prefix1 = 'ACK'
        Domain = 2
        AccHash = 3
        Stream = 4
        Consumer = 5
        NumDelivered = 6
        StreamSeq = 7
        ConsumerSeq = 8
        Timestamp = 9
        NumPending = 10

        # Subject without domain:
        # $JS.ACK.<stream>.<consumer>.<delivered>.<sseq>.<cseq>.<tm>.<pending>
        #
        V1TokenCount = 9

        # Subject with domain:
        # $JS.ACK.<domain>.<account hash>.<stream>.<consumer>.<delivered>.<sseq>.<cseq>.<tm>.<pending>.<a token with a random value>
        #
        V2TokenCount = 12

    def __init__(
        self,
        subject: str = '',
        reply: str = '',
        data: str = b'',
        sid: int = 0,
        client=None,
        headers: dict = None
    ):
        self.subject = subject
        self.reply = reply
        self.data = data
        self.sid = sid
        self._client = client
        self.headers = headers
        self._metadata = None

    @property
    def header(self):
        """
        header returns the headers from a message.
        """
        return self.headers

    async def respond(self, data: bytes):
        """
        respond replies to the inbox of the message if there is one.
        """
        if not self.reply:
            raise Error('no reply subject available')
        if not self._client:
            raise Error('client not set')

        await self._client.publish(self.reply, data, headers=self.headers)

    async def ack(self):
        """
        ack acknowledges a message delivered by JetStream.
        """
        if self.reply is None or self.reply == '':
            raise NotJSMessageError
        await self._client.publish(self.reply)

    async def ack_sync(self, timeout: float = 1.0):
        """
        ack_sync waits for the acknowledgement to be processed by the server.
        """
        if self.reply is None or self.reply == '':
            raise NotJSMessageError

        resp = await self._client.request(self.reply, timeout=timeout)
        return resp

    async def nak(self):
        """
        ack negatively acknowledges a message delivered by JetStream.
        """
        if self.reply is None or self.reply == '':
            raise NotJSMessageError
        await self._client.publish(self.reply, Msg.Ack.Nak)

    async def in_progress(self):
        """
        ack acknowledges a message delivered by JetStream is still being worked on.
        """
        if self.reply is None or self.reply == '':
            raise NotJSMessageError
        await self._client.publish(self.reply, Msg.Ack.Progress)

    async def term(self):
        """
        ack terminates a message delivered by JetStream and disables redeliveries.
        """
        if self.reply is None or self.reply == '':
            raise NotJSMessageError
        await self._client.publish(self.reply, Msg.Ack.Term)

    @property
    def metadata(self):
        msg = self
        # Memoize the parsed metadata.
        metadata = msg._metadata
        if metadata is not None:
            return metadata

        # TODO: Support V2TokenCount style with domains.
        tokens = Msg.Metadata._get_metadata_fields(msg.reply)
        t = datetime.datetime.fromtimestamp(int(tokens[7]) / 1_000_000_000.0)
        metadata = Msg.Metadata(
            sequence=Msg.Metadata.SequencePair(
                stream=int(tokens[5]), consumer=int(tokens[6])
            ),
            num_delivered=int(tokens[4]),
            num_pending=int(tokens[8]),
            timestamp=t,
            stream=tokens[2],
            consumer=tokens[3],
        )
        msg._metadata = metadata
        return metadata

    def _get_metadata_fields(self, reply):
        return Msg.Metadata._get_metadata_fields(reply)

    def __repr__(self):
        return f"<{self.__class__.__name__}: subject='{self.subject}' reply='{self.reply}'>"

    class Metadata:
        """
        Metadata is the metadata from a JetStream message.
        """
        __slots__ = (
            'num_delivered', 'num_pending', 'timestamp', 'stream', 'consumer',
            'sequence'
        )

        @dataclass
        class SequencePair:
            """
            SequencePair represents a pair of consumer and stream sequence.
            """
            consumer: int
            stream: int

        def __init__(
            self,
            sequence=None,
            num_pending: int = None,
            num_delivered: int = None,
            timestamp: str = None,
            stream: str = None,
            consumer: str = None,
        ):
            self.sequence = sequence
            self.num_pending = num_pending
            self.num_delivered = num_delivered
            self.timestamp = timestamp
            self.stream = stream
            self.consumer = consumer

        def _get_metadata_fields(reply):
            if reply is None or reply == '':
                raise NotJSMessageError
            tokens = reply.split('.')
            if len(tokens) != Msg.Ack.V1TokenCount or \
                   tokens[0] != Msg.Ack.Prefix0 or \
                   tokens[1] != Msg.Ack.Prefix1:
                raise NotJSMessageError
            return tokens

        def __repr__(self):
            return f"<{self.__class__.__name__}: stream='{self.stream}' consumer='{self.consumer}' sequence=({self.sequence.stream}, {self.sequence.consumer}) timestamp={self.timestamp}>"



# Copyright 2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio


class Error(Exception):
    pass


class TimeoutError(asyncio.TimeoutError):
    def __str__(self):
        return "nats: timeout"


class ConnectionClosedError(Error):
    def __str__(self):
        return "nats: connection closed"


class SecureConnRequiredError(Error):
    def __str__(self):
        return "nats: secure connection required"


class SecureConnWantedError(Error):
    def __str__(self):
        return "nats: secure connection not available"


class SecureConnFailedError(Error):
    def __str__(self):
        return "nats: secure connection failed"


class BadSubscriptionError(Error):
    def __str__(self):
        return "nats: invalid subscription"


class BadSubjectError(Error):
    def __str__(self):
        return "nats: invalid subject"


class SlowConsumerError(Error):
    def __init__(self, subject=None, sid=None, sub=None):
        self.subject = subject
        self.sid = sid
        self.sub = sub

    def __str__(self):
        return "nats: slow consumer, messages dropped"


class BadTimeoutError(Error):
    def __str__(self):
        return "nats: timeout invalid"


class AuthorizationError(Error):
    def __str__(self):
        return "nats: authorization failed"


class NoServersError(Error):
    def __str__(self):
        return "nats: no servers available for connection"


class JsonParseError(Error):
    def __str__(self):
        return "nats: connect message, json parse err"


class StaleConnectionError(Error):
    def __str__(self):
        return "nats: stale connection"


class MaxPayloadError(Error):
    def __str__(self):
        return "nats: maximum payload exceeded"


class DrainTimeoutError(TimeoutError):
    def __str__(self):
        return "nats: draining connection timed out"


class ConnectionDrainingError(Error):
    def __str__(self):
        return "nats: connection draining"


class ConnectionReconnectingError(Error):
    def __str__(self):
        return "nats: connection reconnecting"


class InvalidUserCredentialsError(Error):
    def __str__(self):
        return "nats: invalid user credentials"


class InvalidCallbackTypeError(Error):
    def __str__(self):
        return "nats: callbacks must be coroutine functions"


class NoRespondersError(Error):
    def __str__(self):
        return "nats: no responders available for request"


class ProtocolError(Error):
    def __str__(self):
        return "nats: protocol error"



# Copyright 2016-2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import nats.errors
from nats.js import api


class Error(nats.errors.Error):
    """
    An Error raised by the NATS client when using JetStream.
    """
    def __init__(self, description=None):
        self.description = description

    def __str__(self):
        desc = ''
        if self.description:
            desc = self.description
        return f"nats: JetStream.{self.__class__.__name__} {desc}"


class APIError(Error):
    """
    An error that is the result of interacting with NATS JetStream.
    """
    def __init__(
        self,
        code=None,
        description=None,
        err_code=None,
        stream=None,
        seq=None
    ):
        self.code = code
        self.err_code = err_code
        self.description = description
        self.stream = stream
        self.seq = seq

    @classmethod
    def from_msg(cls, msg):
        code = msg.header[api.StatusHdr]
        if code == api.ServiceUnavailableStatus:
            raise ServiceUnavailableError
        else:
            desc = msg.header[api.DescHdr]
            raise APIError(code=int(code), description=desc)

    @classmethod
    def from_error(cls, err):
        code = err['code']
        if code == 503:
            raise ServiceUnavailableError(**err)
        elif code == 500:
            raise ServerError(**err)
        elif code == 404:
            raise NotFoundError(**err)
        elif code == 400:
            raise BadRequestError(**err)
        else:
            raise APIError(**err)

    def __str__(self):
        return f"nats: {self.__class__.__name__}: code={self.code} err_code={self.err_code} description='{self.description}'"


class ServiceUnavailableError(APIError):
    """
    A 503 error
    """
    pass


class ServerError(APIError):
    """
    A 500 error
    """
    pass


class NotFoundError(APIError):
    """
    A 404 error
    """
    pass


class BadRequestError(APIError):
    """
    A 400 error
    """
    pass


class NotJSMessageError(Error):
    """
    When it is attempted to use an API meant for JetStream on a message
    xthat does not belong to a stream.
    """
    def __str__(self):
        return "nats: not a JetStream message"


class NoStreamResponseError(Error):
    """
    Raised if the client gets a 503 when publishing a message.
    """
    def __str__(self):
        return "nats: no response from stream"


class ConsumerSequenceMismatchError(Error):
    """
    Async error raised by the client with idle_heartbeat mode enabled
    when one of the message sequences is not the expected one.
    """
    def __init__(
        self,
        stream_resume_sequence=None,
        consumer_sequence=None,
        last_consumer_sequence=None
    ):
        self.stream_resume_sequence = stream_resume_sequence
        self.consumer_sequence = consumer_sequence
        self.last_consumer_sequence = last_consumer_sequence

    def __str__(self):
        gap = self.last_consumer_sequence - self.consumer_sequence
        return f"nats: sequence mismatch for consumer at sequence {self.consumer_sequence} ({gap} sequences behind), should restart consumer from stream sequence {self.stream_resume_sequence}"


class BucketNotFoundError(NotFoundError):
    """
    When attempted to bind to a JetStream KeyValue that does not exist.
    """
    pass


class BadBucketError(Error):
    pass


class KeyDeletedError(Error):
    """
    Raised when trying to get a key that was deleted from a JetStream KeyValue store.
    """
    def __init__(self, entry=None, op=None):
        self.entry = entry
        self.op = op

    def __str__(self):
        return "nats: key was deleted"



# Copyright 2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass, field, fields, asdict
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Dict, List, Optional
import json

MsgIdHdr = "Nats-Msg-Id"
ExpectedStreamHdr = "Nats-Expected-Stream"
ExpectedLastSeqHdr = "Nats-Expected-Last-Sequence"
ExpectedLastSubjSeqHdr = "Nats-Expected-Last-Subject-Sequence"
ExpectedLastMsgIdHdr = "Nats-Expected-Last-Msg-Id"
MsgRollup = "Nats-Rollup"
LastConsumerSeqHdr = "Nats-Last-Consumer"
LastStreamSeqHdr = "Nats-Last-Stream"
StatusHdr = "Status"
DescHdr = "Description"
ServiceUnavailableStatus = "503"
NoMsgsStatus = "404"
CtrlMsgStatus = "100"
DefaultPrefix = "$JS.API"
InboxPrefix = bytearray(b'_INBOX.')


@dataclass
class Base:
    """
    Helper dataclass to filter unknown fields from the API.
    """
    @classmethod
    def properties(klass, **opts):
        return [f.name for f in fields(klass)]

    @classmethod
    def loads(klass, **opts):
        # Reject unknown properties before loading.
        # FIXME: Find something more efficient...
        to_rm = []
        for e in opts:
            if e not in klass.properties():
                to_rm.append(e)

        for m in to_rm:
            del opts[m]
        return klass(**opts)

    def asjson(self):
        # Filter and remove any null values since invalid for Go.
        cfg = asdict(self)
        for k, v in dict(cfg).items():
            if v is None:
                del cfg[k]
        return json.dumps(cfg)


@dataclass
class PubAck(Base):
    """
    PubAck is the response of publishing a message to JetStream.
    """
    stream: str
    seq: int
    domain: Optional[str] = None
    duplicate: Optional[bool] = None


@dataclass
class Placement(Base):
    """Placement directives to consider when placing replicas of this stream"""

    cluster: str
    tags: Optional[List[str]] = None


@dataclass
class ExternalStream(Base):
    api: str
    deliver: Optional[str] = None


@dataclass
class StreamSource(Base):
    name: str
    opt_start_seq: Optional[int] = None
    # FIXME: Handle time type, omit for now.
    # opt_start_time: Optional[str] = None
    filter_subject: Optional[str] = None
    external: Optional[ExternalStream] = None

    def __post_init__(self):
        if isinstance(self.external, dict):
            self.external = ExternalStream.loads(**self.external)


@dataclass
class StreamSourceInfo(Base):
    name: str
    lag: Optional[int] = None
    active: Optional[int] = None
    error: Optional[Dict[str, Any]] = None


@dataclass
class LostStreamData(Base):
    msgs: Optional[List[int]] = None
    bytes: Optional[int] = None


@dataclass
class StreamState(Base):
    messages: int
    bytes: int
    first_seq: int
    last_seq: int
    consumer_count: int
    deleted: Optional[List[int]] = None
    num_deleted: Optional[int] = None
    lost: Optional[LostStreamData] = None

    def __post_init__(self):
        if isinstance(self.lost, dict):
            self.lost = LostStreamData.loads(**self.lost)


class RetentionPolicy(str, Enum):
    """How message retention is considered"""

    limits = "limits"
    interest = "interest"
    workqueue = "workqueue"


class StorageType(str, Enum):
    """The type of storage backend"""

    file = "file"
    memory = "memory"


class DiscardPolicy(str, Enum):
    """Discard policy when a stream reaches its limits"""

    old = "old"
    new = "new"


@dataclass
class StreamConfig(Base):
    """
    StreamConfig represents the configuration of a stream.
    """
    name: Optional[str] = None
    description: Optional[str] = None
    subjects: Optional[List[str]] = None
    retention: Optional[RetentionPolicy] = None
    max_consumers: Optional[int] = None
    max_msgs: Optional[int] = None
    max_bytes: Optional[int] = None
    discard: Optional[DiscardPolicy] = DiscardPolicy.old
    max_age: Optional[int] = None
    max_msgs_per_subject: Optional[int] = -1
    max_msg_size: Optional[int] = -1
    storage: Optional[StorageType] = None
    num_replicas: Optional[int] = None
    no_ack: Optional[bool] = False
    template_owner: Optional[str] = None
    duplicate_window: Optional[int] = 0
    placement: Optional[Placement] = None
    mirror: Optional[StreamSource] = None
    sources: Optional[List[StreamSource]] = None
    sealed: Optional[bool] = False
    deny_delete: Optional[bool] = False
    deny_purge: Optional[bool] = False
    allow_rollup_hdrs: Optional[bool] = False

    def __post_init__(self):
        if isinstance(self.placement, dict):
            self.placement = Placement.loads(**self.placement)
        if isinstance(self.mirror, dict):
            self.mirror = StreamSource.loads(**self.mirror)
        if self.sources:
            self.sources = [
                StreamSource.loads(**item) if isinstance(item, dict) else item
                for item in self.sources
            ]


@dataclass
class PeerInfo(Base):
    name: Optional[str] = None
    current: Optional[bool] = None
    offline: Optional[bool] = None
    active: Optional[int] = None
    lag: Optional[int] = None


@dataclass
class ClusterInfo(Base):
    leader: Optional[str] = None
    name: Optional[str] = None
    replicas: Optional[List[PeerInfo]] = None

    def __post_init__(self):
        if self.replicas:
            self.replicas = [
                PeerInfo.loads(**item) if isinstance(item, dict) else item
                for item in self.replicas
            ]


@dataclass
class StreamInfo(Base):
    """
    StreamInfo is the latest information about a stream from JetStream.
    """
    config: StreamConfig
    state: StreamState
    mirror: Optional[StreamSourceInfo] = None
    sources: Optional[List[StreamSourceInfo]] = None
    cluster: Optional[ClusterInfo] = None
    did_create: Optional[bool] = None

    def __post_init__(self):
        if isinstance(self.config, dict):
            self.config = StreamConfig.loads(**self.config)
        if isinstance(self.state, dict):
            self.state = StreamState.loads(**self.state)
        if isinstance(self.mirror, dict):
            self.mirror = StreamSourceInfo.loads(**self.mirror)
        if self.sources:
            self.sources = [
                StreamSourceInfo.loads(**item)
                if isinstance(item, dict) else item for item in self.sources
            ]
        if isinstance(self.cluster, dict):
            self.cluster = ClusterInfo.loads(**self.cluster)


class AckPolicy(str, Enum):
    """Policies defining how messages should be acknowledged.

    If an ack is required but is not received within the AckWait window, the message will be redelivered.

    References:
        * `Consumers, AckPolicy <https://docs.nats.io/jetstream/concepts/consumers#ackpolicy>`_
    """

    none = "none"
    all = "all"
    explicit = "explicit"


class DeliverPolicy(str, Enum):
    """When a consumer is first created, it can specify where in the stream it wants to start receiving messages.

    This is the DeliverPolicy, and this enumeration defines allowed values.

    References:
        * `Consumers, DeliverPolicy/OptStartSeq/OptStartTime <https://docs.nats.io/jetstream/concepts/consumers#deliverpolicy-optstartseq-optstarttime>`_
    """

    all = "all"
    last = "last"
    new = "new"
    last_per_subject = "last_per_subject"
    by_start_sequence = "by_start_sequence"
    by_start_time = "by_start_time"


class ReplayPolicy(str, Enum):
    """The replay policy applies when the DeliverPolicy is one of:
        * all
        * by_start_sequence
        * by_start_time
    since those deliver policies begin reading the stream at a position other than the end.

    References:
        * `Consumers, ReplayPolicy <https://docs.nats.io/jetstream/concepts/consumers#replaypolicy>`_
    """

    instant = "instant"
    original = "original"


@dataclass
class ConsumerConfig(Base):
    """Consumer configuration.

    References:
        * `Consumers <https://docs.nats.io/jetstream/concepts/consumers>`_
    """
    durable_name: Optional[str] = None
    description: Optional[str] = None
    deliver_subject: Optional[str] = None
    deliver_group: Optional[str] = None
    deliver_policy: Optional[DeliverPolicy] = DeliverPolicy.all
    opt_start_seq: Optional[int] = None
    opt_start_time: Optional[int] = None
    ack_policy: Optional[AckPolicy] = AckPolicy.explicit
    ack_wait: Optional[int] = None
    max_deliver: Optional[int] = None
    filter_subject: Optional[str] = None
    replay_policy: Optional[ReplayPolicy] = ReplayPolicy.instant
    sample_freq: Optional[str] = None
    rate_limit_bps: Optional[int] = None
    max_waiting: Optional[int] = None
    max_ack_pending: Optional[int] = None
    flow_control: Optional[bool] = None
    idle_heartbeat: Optional[int] = None
    headers_only: Optional[bool] = None


@dataclass
class SequenceInfo(Base):
    consumer_seq: int
    stream_seq: int
    # FIXME: Do not handle dates for now.
    # last_active: Optional[datetime]


@dataclass
class ConsumerInfo(Base):
    """
    ConsumerInfo represents the info about the consumer.
    """
    stream_name: str
    config: ConsumerConfig
    # FIXME: Do not handle dates for now.
    # created: datetime
    delivered: Optional[SequenceInfo] = None
    ack_floor: Optional[SequenceInfo] = None
    num_ack_pending: Optional[int] = None
    num_redelivered: Optional[int] = None
    num_waiting: Optional[int] = None
    num_pending: Optional[int] = None
    name: Optional[str] = None
    cluster: Optional[ClusterInfo] = None
    push_bound: Optional[bool] = None

    def __post_init__(self):
        if isinstance(self.delivered, dict):
            self.delivered = SequenceInfo.loads(**self.delivered)
        if isinstance(self.ack_floor, dict):
            self.ack_floor = SequenceInfo.loads(**self.ack_floor)
        if isinstance(self.config, dict):
            self.config = ConsumerConfig.loads(**self.config)
        if isinstance(self.cluster, dict):
            self.cluster = ClusterInfo.loads(**self.cluster)


@dataclass
class AccountLimits(Base):
    """Account limits

    References:
        * `Multi-tenancy & Resource Mgmt <https://docs.nats.io/jetstream/resource_management>`_
    """

    max_memory: int
    max_storage: int
    max_streams: int
    max_consumers: int


@dataclass
class APIStats(Base):
    """API stats"""

    total: int
    errors: int


@dataclass
class AccountInfo(Base):
    """Account information

    References:
        * `Account Information <https://docs.nats.io/jetstream/administration/account#account-information>`_
    """

    memory: int
    storage: int
    streams: int
    consumers: int
    limits: AccountLimits
    api: APIStats
    domain: Optional[str] = None

    def __post_init__(self):
        if isinstance(self.limits, dict):
            self.limits = AccountLimits.loads(**self.limits)
        if isinstance(self.api, dict):
            self.api = APIStats.loads(**self.api)


@dataclass
class RawStreamMsg(Base):
    subject: str = None
    seq: int = None
    data: bytes = None
    hdrs: bytes = None
    headers: dict = None
    # TODO: Add 'time'

    @property
    def sequence(self):
        return self.seq


@dataclass
class KeyValueConfig(Base):
    """
    KeyValueConfig is the configurigation of a KeyValue store.
    """
    bucket: Optional[str] = None
    description: Optional[str] = None
    max_value_size: Optional[int] = None
    history: Optional[int] = None
    ttl: int = None  # in seconds
    max_bytes: Optional[int] = None
    storage: Optional[StorageType] = None
    replicas: Optional[int] = None



# Copyright 2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from nats.js import api
from nats.errors import *
from nats.js.errors import *
from dataclasses import dataclass, asdict
from typing import Any, Dict, List, Optional
import base64

KV_OP = "KV-Operation"
KV_DEL = "DEL"
KV_PURGE = "PURGE"
KV_EXPECTED_HDR = "Nats-Expected-Last-Subject-Sequence"
MSG_ID_HDR = "Nats-Msg-Id"
EXPECTED_STREAM_HDR = "Nats-Expected-Stream"
EXPECTED_LAST_SEQ_HDR = "Nats-Expected-Last-Sequence"
EXPECTED_LAST_SUBJ_SEQ_HDR = "Nats-Expected-Last-Subject-Sequence"
EXPECTED_LAST_MSG_ID_HDR = "Nats-Expected-Last-Msg-Id"
MSG_ROLLUP = "Nats-Rollup"
MSG_ROLLUP_SUBJECT = "sub"
MSG_ROLLUP_ALL = "all"


class KeyValue:
    """
    KeyValue uses the JetStream KeyValue functionality.

    .. note::
       This functionality is EXPERIMENTAL and may be changed in later releases.   

    ::

        import asyncio
        import nats

        async def main():
            nc = await nats.connect()
            js = nc.jetstream()

            # Create a KV
            kv = await js.create_key_value(bucket='MY_KV')

            # Set and retrieve a value
            await kv.put('hello', b'world')
            entry = await kv.get('hello')
            print(f'KeyValue.Entry: key={entry.key}, value={entry.value}')
            # KeyValue.Entry: key=hello, value=world

            await nc.close()

        if __name__ == '__main__':
            asyncio.run(main())

    """
    @dataclass
    class Entry:
        """
        An entry from a KeyValue store in JetStream.       
        """
        bucket: str
        key: str
        value: bytes
        revision: int

    class BucketStatus:
        """
        BucketStatus is the status of a KeyValue bucket.
        """
        def __init__(self):
            self._nfo = None
            self._bucket = None

        @property
        def bucket(self):
            """
            name of the bucket the key value.
            """
            return self._bucket

        @property
        def values(self):
            """
            values returns the number of stored messages in the stream.
            """
            return self._nfo.state.messages

        @property
        def history(self):
            """
            history returns the max msgs per subject.
            """
            return self._nfo.config.max_msgs_per_subject

        @property
        def ttl(self):
            """
            ttl returns the max age in seconds.
            """
            return self._nfo.config.max_age / 1_000_000_000

        @property
        def stream_info(self):
            return self._nfo

        def __repr__(self):
            return f"<KeyValue.{self.__class__.__name__}: bucket={self.bucket} values={self.values} history={self.history} ttl={self.ttl}>"

    def __init__(self):
        self._name = None
        self._stream = None
        self._pre = None
        self._js = None

    async def get(self, key: str) -> Entry:
        """
        get returns the latest value for the key.
        """
        msg = await self._js.get_last_msg(self._stream, f"{self._pre}{key}")
        data = None
        if msg.data:
            data = base64.b64decode(msg.data)

        entry = KeyValue.Entry(
            bucket=self._name,
            key=key,
            value=data,
            revision=msg.seq,
        )

        # Check headers to see if deleted or purged.
        if msg.headers:
            op = msg.headers.get(KV_OP, None)
            if op == KV_DEL or op == KV_PURGE:
                raise KeyDeletedError(entry, op)

        return entry

    async def put(self, key: str, value: bytes) -> int:
        """
        put will place the new value for the key into the store
        and return the revision number.
        """
        pa = await self._js.publish(f"{self._pre}{key}", value)
        return pa.seq

    async def update(self, key: str, value: bytes, last: int) -> int:
        """
        update will update the value iff the latest revision matches.
        """
        hdrs = {}
        hdrs[KV_EXPECTED_HDR] = str(last)
        pa = await self._js.publish(f"{self._pre}{key}", value, headers=hdrs)
        return pa.sequence

    async def delete(self, key: str):
        """
        delete will place a delete marker and remove all previous revisions.
        """
        hdrs = {}
        hdrs[KV_OP] = KV_DEL
        await self._js.publish(f"{self._pre}{key}", headers=hdrs)
        return True

    async def purge(self, key: str):
        """
        purge will remove the key and all revisions.
        """
        hdrs = {}
        hdrs[KV_OP] = KV_PURGE
        hdrs[MSG_ROLLUP] = MSG_ROLLUP_SUBJECT
        await self._js.publish(f"{self._pre}{key}", headers=hdrs)
        return True

    async def status(self) -> BucketStatus:
        """
        status retrieves the status and configuration of a bucket.
        """
        info = await self._js.stream_info(self._stream)
        kvs = KeyValue.BucketStatus()
        kvs._nfo = info
        kvs._bucket = self._name
        return kvs


class KeyValueManager:
    """
    KeyValueManager includes methods to create KV stores in JetStream.
    """
    async def key_value(self, bucket: str):
        stream = f"KV_{bucket}"
        try:
            si = await self.stream_info(stream)
        except NotFoundError:
            raise BucketNotFoundError
        if si.config.max_msgs_per_subject < 1:
            raise BadBucketError

        kv = KeyValue()
        kv._name = bucket
        kv._stream = stream
        kv._pre = f"$KV.{bucket}."
        kv._js = self
        return kv

    async def create_key_value(self, **config) -> KeyValue:
        """
        create_key_value takes an api.KeyValueConfig and creates a KV in JetStream.
        """
        if not isinstance(config, api.KeyValueConfig):
            config = api.KeyValueConfig.loads(**config)

        if not config.history:
            config.history = 1
        if not config.replicas:
            config.replicas = 1
        if config.ttl:
            config.ttl = config.ttl * 1_000_000_000

        stream = api.StreamConfig(
            name=f"KV_{config.bucket}",
            description=None,
            subjects=[f"$KV.{config.bucket}.>"],
            max_msgs_per_subject=config.history,
            max_bytes=config.max_bytes,
            max_age=config.ttl,
            max_msg_size=config.max_value_size,
            storage=config.storage,
            num_replicas=config.replicas,
            allow_rollup_hdrs=True,
            deny_delete=True,
        )
        resp = await self.add_stream(stream)

        kv = KeyValue()
        kv._name = config.bucket
        kv._stream = stream.name
        kv._pre = f"$KV.{config.bucket}."
        kv._js = self
        return kv

    async def delete_key_value(self, bucket: str):
        """
        delete_key_value deletes a JetStream KeyValue store by destroying
        the associated stream.
        """
        result = await self._js.delete_stream(self._stream)
        return result



# Copyright 2016-2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import nats.errors


class NatsError(nats.errors.Error):
    """
    .. deprecated:: v2.0.0


    Please use `nats.errors.Error` instead.
    """
    pass


class ErrConnectionClosed(nats.errors.ConnectionClosedError):
    """
    
    .. deprecated:: v2.0.0

    Please use `nats.errors.ConnectionClosedError` instead.
    """
    pass


class ErrDrainTimeout(nats.errors.DrainTimeoutError):
    """
    
    .. deprecated:: v2.0.0

    Please use `nats.errors.DrainTimeoutError` instead.
    """
    pass


class ErrInvalidUserCredentials(nats.errors.InvalidUserCredentialsError):
    """
    
    .. deprecated:: v2.0.0

    Please use `nats.errors.InvalidUserCredentialsError` instead.
    """
    pass


class ErrInvalidCallbackType(nats.errors.InvalidCallbackTypeError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.InvalidCallbackTypeError` instead.
    """
    pass


class ErrConnectionReconnecting(nats.errors.ConnectionReconnectingError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.ConnectionReconnectingError` instead.
    """
    pass


class ErrConnectionDraining(nats.errors.ConnectionDrainingError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.ConnectionDrainingError` instead.
    """
    pass


class ErrMaxPayload(nats.errors.MaxPayloadError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.MaxPayloadError` instead.
    """
    pass


class ErrStaleConnection(nats.errors.StaleConnectionError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.StaleConnectionError` instead.
    """
    pass


class ErrJsonParse(nats.errors.JsonParseError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.JsonParseError` instead.
    """
    pass


class ErrSecureConnRequired(nats.errors.SecureConnRequiredError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.SecureConnRequiredError` instead.
    """
    pass


class ErrSecureConnWanted(nats.errors.SecureConnWantedError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.SecureConnWantedError` instead.
    """
    pass


class ErrSecureConnFailed(nats.errors.SecureConnFailedError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.SecureConnFailedError` instead.
    """
    pass


class ErrBadSubscription(nats.errors.BadSubscriptionError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.BadSubscriptionError` instead.
    """
    pass


class ErrBadSubject(nats.errors.BadSubjectError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.BadSubjectError` instead.
    """
    pass


class ErrSlowConsumer(nats.errors.SlowConsumerError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.SlowConsumerError` instead.
    """
    pass


class ErrTimeout(nats.errors.TimeoutError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.TimeoutError` instead.
    """
    pass


class ErrBadTimeout(nats.errors.BadTimeoutError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.BadTimeoutError` instead.
    """
    pass


class ErrAuthorization(nats.errors.AuthorizationError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.AuthorizationError` instead.
    """
    pass


class ErrNoServers(nats.errors.NoServersError):
    """

    .. deprecated:: v2.0.0

    Please use `nats.errors.NoServersError` instead.
    """
    pass



# Copyright 2021 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import json
import time
import nats.errors
import nats.js.errors
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription
from nats.js.manager import JetStreamManager
from nats.js.kv import KeyValueManager
from nats.js import api
from typing import Any, Dict, List, Optional, Callable
from dataclasses import asdict

LAST_CONSUMER_SEQ_HDR = "Nats-Last-Consumer"


class JetStream:
    """
    JetStream returns a context that can be used to produce and consume
    messages from NATS JetStream.

    :param conn: NATS Connection
    :param prefix: Default JetStream API Prefix.
    :param domain: Optional domain used by the JetStream API.
    :param timeout: Timeout for all JS API actions.

    ::

        import asyncio
        import nats

        async def main():
            nc = await nats.connect()
            js = nc.jetstream()

            await js.add_stream(name='hello', subjects=['hello'])
            ack = await js.publish('hello', b'Hello JS!')
            print(f'Ack: stream={ack.stream}, sequence={ack.seq}')
            # Ack: stream=hello, sequence=1
            await nc.close()

        if __name__ == '__main__':
            asyncio.run(main())

    """
    def __init__(
        self,
        conn,
        prefix=api.DefaultPrefix,
        domain=None,
        timeout=5,
    ):
        self._prefix = prefix
        if domain is not None:
            self._prefix = f"$JS.{domain}.API"
        self._nc = conn
        self._timeout = timeout
        self._jsm = JetStreamManager(
            conn, prefix=prefix, domain=domain, timeout=timeout
        )

    async def publish(
        self,
        subject: str,
        payload: bytes = b'',
        timeout: float = None,
        stream: str = None,
        headers: dict = None
    ) -> api.PubAck:
        """
        publish emits a new message to JetStream.
        """
        hdr = headers
        if timeout is None:
            timeout = self._timeout
        if stream is not None:
            if headers is None:
                hdr = {}
            hdr[nats.js.api.ExpectedStreamHdr] = stream

        try:
            msg = await self._nc.request(
                subject, payload, timeout=timeout, headers=hdr
            )
        except nats.errors.NoRespondersError:
            raise nats.js.errors.NoStreamResponseError

        resp = json.loads(msg.data)
        if 'error' in resp:
            raise nats.js.errors.APIError.from_error(resp['error'])

        return api.PubAck.loads(**resp)

    async def subscribe(
        self,
        subject: str,
        queue: Optional[str] = None,
        cb: Optional[Callable[['Msg'], None]] = None,
        durable: Optional[str] = None,
        stream: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        manual_ack: Optional[bool] = False,
        ordered_consumer: Optional[bool] = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: Optional[bool] = False,
    ):
        """
        subscribe returns a `Subscription` that is bound to a push based consumer.

        :param subject: Subject from a stream from JetStream.
        :param queue: Deliver group name from a set a of queue subscribers.
        :param durable: Name of the durable consumer to which the the subscription should be bound.
        :param stream: Name of the stream to which the subscription should be bound. If not set,
          then the client will automatically look it up based on the subject.
        :param manual_ack: Disables auto acking for async subscriptions.
        :param ordered_consumer: Enable ordered consumer mode.
        :param idle_heartbeat: Enable Heartbeats for a consumer to detect failures.
        :param flow_control: Enable Flow Control for a consumer.

        ::

            import asyncio
            import nats

            async def main():
                nc = await nats.connect()
                js = nc.jetstream()

                await js.add_stream(name='hello', subjects=['hello'])
                await js.publish('hello', b'Hello JS!')

                async def cb(msg):
                  print('Received:', msg)

                # Ephemeral Async Subscribe
                await js.subscribe('hello', cb=cb)

                # Durable Async Subscribe
                # NOTE: Only one subscription can be bound to a durable name. It also auto acks by default.
                await js.subscribe('hello', cb=cb, durable='foo')

                # Durable Sync Subscribe
                # NOTE: Sync subscribers do not auto ack.
                await js.subscribe('hello', durable='bar')

                # Queue Async Subscribe
                # NOTE: Here 'workers' becomes deliver_group, durable name and queue name.
                await js.subscribe('hello', 'workers', cb=cb)

            if __name__ == '__main__':
                asyncio.run(main())

        """
        if stream is None:
            stream = await self._jsm.find_stream_name_by_subject(subject)

        deliver = None
        consumer = None

        # If using a queue, that will be the consumer/durable name.
        if queue:
            if durable and durable != queue:
                raise nats.js.errors.Error(
                    f"cannot create queue subscription '{queue}' to consumer '{durable}'"
                )
            else:
                durable = queue

        cinfo = None
        consumerFound = False
        shouldCreate = False

        # Ephemeral subscribe always has to be auto created.
        if not durable:
            shouldCreate = True
        else:
            try:
                # TODO: Detect configuration drift with any present durable consumer.
                cinfo = await self._jsm.consumer_info(stream, durable)
                config = cinfo.config
                consumerFound = True
                consumer = durable
            except nats.js.errors.NotFoundError:
                shouldCreate = True
                consumerFound = False

        if consumerFound:
            # At this point, we know the user wants push mode, and the JS consumer is
            # really push mode.
            if not config.deliver_group:
                # Prevent an user from attempting to create a queue subscription on
                # a JS consumer that was not created with a deliver group.
                if queue:
                    # TODO: Currently, this would not happen in client
                    # since the queue name is used as durable name.
                    raise nats.js.errors.Error(
                        "cannot create a queue subscription for a consumer without a deliver group"
                    )
                elif cinfo.push_bound:
                    # Need to reject a non queue subscription to a non queue consumer
                    # if the consumer is already bound.
                    raise nats.js.errors.Error(
                        "consumer is already bound to a subscription"
                    )
            else:
                if not queue:
                    raise nats.js.errors.Error(
                        f"cannot create a subscription for a consumer with a deliver group {config.deliver_group}"
                    )
                elif queue != config.deliver_group:
                    raise nats.js.errors.Error(
                        f"cannot create a queue subscription {queue} for a consumer with a deliver group {config.deliver_group}"
                    )
        elif shouldCreate:
            # Auto-create consumer if none found.
            if config is None:
                # Defaults
                config = api.ConsumerConfig(
                    ack_policy=api.AckPolicy.explicit,
                )
            elif isinstance(config, dict):
                config = api.ConsumerConfig.loads(**config)
            elif not isinstance(config, api.ConsumerConfig):
                raise ValueError("nats: invalid ConsumerConfig")

            if not config.durable_name:
                config.durable_name = durable
            if not config.deliver_group:
                config.deliver_group = queue

            # Create inbox for push consumer.
            deliver = self._nc.new_inbox()
            config.deliver_subject = deliver

            # Auto created consumers use the filter subject.
            config.filter_subject = subject

            # Heartbeats / FlowControl
            config.flow_control = flow_control
            if idle_heartbeat or config.idle_heartbeat:
                if config.idle_heartbeat:
                    idle_heartbeat = config.idle_heartbeat
                idle_heartbeat = int(idle_heartbeat * 1_000_000_000)
                config.idle_heartbeat = idle_heartbeat

            # Enable ordered consumer mode where at most there is
            # one message being delivered at a time.
            if ordered_consumer:
                if not idle_heartbeat:
                    idle_heartbeat = 5 * 1_000_000_000
                config.flow_control = True
                config.ack_policy = api.AckPolicy.explicit
                config.max_deliver = 1
                config.ack_wait = 22 * 3600 * 1_000_000_000  # 22 hours
                config.idle_heartbeat = idle_heartbeat

            cinfo = await self._jsm.add_consumer(stream, config=config)
            consumer = cinfo.name

        # By default, async subscribers wrap the original callback and
        # auto ack the messages as they are delivered.
        if cb and not manual_ack:
            ocb = cb

            async def new_cb(msg):
                await ocb(msg)
                await msg.ack()

            cb = new_cb

        sub = await self._nc.subscribe(
            config.deliver_subject, config.deliver_group, cb=cb
        )
        psub = JetStream.PushSubscription(self, sub, stream, consumer)

        # Keep state to support ordered consumers.
        jsi = JetStream._JSI()
        jsi._js = self
        jsi._conn = self._nc
        jsi._stream = stream
        jsi._ordered = ordered_consumer
        jsi._psub = psub
        jsi._sub = sub
        jsi._ccreq = config
        sub._jsi = jsi
        return psub

    async def pull_subscribe(
        self,
        subject: str,
        durable: str,
        stream: str = None,
        config: api.ConsumerConfig = None,
    ):
        """
        pull_subscribe returns a `PullSubscription` that can be delivered messages
        from a JetStream pull based consumer by calling `sub.fetch`.

        In case 'stream' is passed, there will not be a lookup of the stream
        based on the subject.

        ::

            import asyncio
            import nats

            async def main():
                nc = await nats.connect()
                js = nc.jetstream()

                await js.add_stream(name='mystream', subjects=['foo'])
                await js.publish('foo', b'Hello World!')

                msgs = await sub.fetch()
                msg = msgs[0]
                await msg.ack()

                await nc.close()

            if __name__ == '__main__':
                asyncio.run(main())

        """
        if stream is None:
            stream = await self._jsm.find_stream_name_by_subject(subject)

        try:
            # TODO: Detect configuration drift with the consumer.
            await self._jsm.consumer_info(stream, durable)
        except nats.js.errors.NotFoundError:
            # If not found then attempt to create with the defaults.
            if config is None:
                # Defaults
                config = api.ConsumerConfig(
                    ack_policy=api.AckPolicy.explicit,
                )
            elif isinstance(config, dict):
                config = api.ConsumerConfig.loads(**config)
            elif not isinstance(config, api.ConsumerConfig):
                raise ValueError("nats: invalid ConsumerConfig")

            # Auto created consumers use the filter subject.
            config.filter_subject = subject
            config.durable_name = durable
            await self._jsm.add_consumer(stream, config=config)

        # FIXME: Make this inbox prefix customizable.
        deliver = api.InboxPrefix[:]
        deliver.extend(self._nc._nuid.next())

        consumer = durable
        sub = await self._nc.subscribe(deliver.decode())
        return JetStream.PullSubscription(self, sub, stream, consumer, deliver)

    @classmethod
    def is_status_msg(cls, msg):
        if msg is not None and \
           msg.header is not None and \
           api.StatusHdr in msg.header:
            return True
        else:
            return False

    class _JSI():
        def __init__(self):
            self._stream = None
            self._ordered = None
            self._conn = None
            self._psub = None
            self._sub = None
            self._dseq = 1
            self._sseq = 0
            self._ccreq = None
            self._cmeta = None
            self._fciseq = 0
            self._active = None

            # flow control response
            self._fcr = None
            # flow control sequence
            self._fcd = None

            # background task that resets an ordered consumer
            self._reset_task = None

        def track_sequences(self, reply):
            self._fciseq += 1
            self._cmeta = reply

        def schedule_flow_control_response(self, reply):
            self._fcr = reply
            self._fcd = self._fciseq

        async def check_for_sequence_mismatch(self, msg):
            self._active = True
            if not self._cmeta:
                return

            tokens = msg._get_metadata_fields(self._cmeta)
            dseq = int(tokens[6])  # consumer sequence
            ldseq = int(msg.header.get(LAST_CONSUMER_SEQ_HDR))
            did_reset = None

            if ldseq != dseq:
                sseq = int(tokens[5])  # stream sequence

                if self._ordered:
                    did_reset = await self.reset_ordered_consumer(
                        self._sseq + 1
                    )
                else:
                    ecs = nats.js.errors.ConsumerSequenceMismatchError(
                        sseq,
                        dseq,
                        ldseq,
                    )
                    if self._conn._error_cb:
                        await self._conn._error_cb(ecs)
            return did_reset

        async def reset_ordered_consumer(self, sseq):
            # FIXME: Handle AUTO_UNSUB called previously to this.

            # Replace current subscription.
            osid = self._sub._id
            self._conn._remove_sub(osid)
            new_deliver = self._conn.new_inbox()

            # Place new one.
            self._conn._sid += 1
            nsid = self._conn._sid
            self._conn._subs[nsid] = self._sub
            self._sub._id = nsid
            self._psub._id = nsid

            # unsub
            await self._conn._send_unsubscribe(osid)

            # resub
            self._sub._subject = new_deliver
            await self._conn._send_subscribe(self._sub)

            # relinquish cpu to let proto commands make it to the server.
            await asyncio.sleep(0)

            # Reset some items in jsi.
            self._cmeta = None
            self._dseq = 1
            self._fcr = None
            self._fcd = None

            # Reset consumer request for starting policy.
            config = self._ccreq
            config.deliver_subject = new_deliver
            config.deliver_policy = nats.js.api.DeliverPolicy.by_start_sequence
            config.opt_start_seq = sseq
            self._ccreq = config

            # Handle the creation of new consumer in a background task
            # to avoid blocking the process_msg coroutine further
            # when making the request.
            self._reset_task = asyncio.create_task(self.recreate_consumer())

            return True

        async def recreate_consumer(self):
            try:
                cinfo = await self._js._jsm.add_consumer(
                    self._stream,
                    config=self._ccreq,
                    timeout=self._js._timeout
                )
                self._psub._consumer = cinfo.name
            except Exception as err:
                await self._conn._error_cb(err)

    class PushSubscription(Subscription):
        """
        PushSubscription is a subscription that is delivered messages.
        """
        def __init__(self, js, sub, stream, consumer):
            self._js = js
            self._stream = stream
            self._consumer = consumer

            self._conn = sub._conn
            self._id = sub._id
            self._subject = sub._subject
            self._queue = sub._queue
            self._max_msgs = sub._max_msgs
            self._received = sub._received
            self._cb = sub._cb
            self._future = sub._future
            self._closed = sub._closed

            # Per subscription message processor.
            self._pending_msgs_limit = sub._pending_msgs_limit
            self._pending_bytes_limit = sub._pending_bytes_limit
            self._pending_queue = sub._pending_queue
            self._pending_size = sub._pending_size
            self._wait_for_msgs_task = sub._wait_for_msgs_task
            self._message_iterator = sub._message_iterator

        async def consumer_info(self):
            """
            consumer_info gets the current info of the consumer from this subscription.
            """
            info = await self._js._jsm.consumer_info(
                self._stream, self._consumer
            )
            return info

    class PullSubscription:
        """
        PullSubscription is a subscription that can fetch messages.
        """
        def __init__(self, js, sub, stream, consumer, deliver):
            # JS/JSM context
            self._js = js
            self._nc = js._nc

            # NATS Subscription
            self._sub = sub
            self._stream = stream
            self._consumer = consumer
            prefix = self._js._prefix
            self._nms = f'{prefix}.CONSUMER.MSG.NEXT.{stream}.{consumer}'
            self._deliver = deliver.decode()

        async def unsubscribe():
            """
            unsubscribe destroys de inboxes of the pull subscription making it
            unable to continue to receive messages.
            """
            if self._sub is None:
                raise ValueError("nats: invalid subscription")

            await self._sub.unsubscribe()
            self._sub = None

        async def fetch(self, batch: int = 1, timeout: int = 5):
            if self._sub is None:
                raise ValueError("nats: invalid subscription")

            # FIXME: Check connection is not closed, etc...

            if batch < 1:
                raise ValueError("nats: invalid batch size")
            if not timeout or timeout <= 0:
                raise ValueError("nats: invalid fetch timeout")

            msgs = []
            expires = (timeout * 1_000_000_000) - 100_000
            if batch == 1:
                msg = await self._fetch_one(batch, expires, timeout)
                msgs.append(msg)
            else:
                msgs = await self._fetch_n(batch, expires, timeout)
            return msgs

        async def _fetch_one(self, batch: int, expires: int, timeout: int):
            queue = self._sub._pending_queue

            # Check the next message in case there are any.
            if not queue.empty():
                try:
                    msg = queue.get_nowait()
                    return msg
                except:
                    # Fallthrough to make request in case this fails.
                    pass

            # Make lingering request with expiration.
            next_req = {}
            next_req['batch'] = 1
            next_req['expires'] = expires

            # Make publish request and wait for response.
            await self._nc.publish(
                self._nms,
                json.dumps(next_req).encode(),
                self._deliver,
            )

            # Wait for the response.
            msg = None
            try:
                fut = queue.get()
                msg = await asyncio.wait_for(fut, timeout=timeout)
            except asyncio.TimeoutError:
                raise nats.errors.TimeoutError

            # Should have received at least a message at this point,
            # if that is not the case then error already.
            if JetStream.is_status_msg(msg):
                if api.StatusHdr in msg.headers:
                    raise nats.js.errors.APIError.from_msg(msg)

            return msg

        async def _fetch_n(self, batch, expires, timeout):
            # TODO: Implement fetching more than one.
            raise NotImplementedError

    class _JS():
        def __init__(
            self,
            conn=None,
            prefix=None,
            stream=None,
            consumer=None,
            nms=None,
        ):
            self._prefix = prefix
            self._nc = conn
            self._stream = stream
            self._consumer = consumer
            self._nms = nms


class JetStreamContext(JetStream, JetStreamManager, KeyValueManager):
    """
    JetStreamContext includes the fully featured context for interacting
    with JetStream.
    """
    def __init__(self, conn, **opts):
        super().__init__(conn, **opts)
